use16
org 0

OLD_INT19H   equ 85h
ROM_SIZE     equ 64 ;K
KEY_WAIT_SEC equ 3 ;seconds
SIGNATURE1   db  055h
SIGNATURE2   db  0AAh
SIGNATURE3   db  ROM_SIZE * 2
jmp short install_int19h

CHECKSUM     db 0
INTERVAL     db 'Press F1 to enter SETUP...',0

install_int19h:
  cli
  push ds
  push ax
  push bx
  xor ax,ax
  mov ds,ax
  mov ax,word[ds:19h*4]
  mov bx,word[ds:19h*4+2]
  mov word[ds:OLD_INT19H*4],ax
  mov word[ds:OLD_INT19H*4+2],bx
  mov ax,entry_point
  mov bx,cs
  mov word[ds:19h*4],ax
  mov word[ds:19h*4+2],bx
  pop bx
  pop ax
  pop ds
  sti
  retf

entry_point:
  cli
  pop ax
  pop ax
  pop ax
  push cs
  mov si,start
  push si
  xor ax,ax
  mov ds,ax
  mov ax,word[ds:OLD_INT19H*4]
  mov bx,word[ds:OLD_INT19H*4+2]
  mov word[ds:19h*4],ax
  mov word[ds:19h*4+2],bx  
  iret
  
start:
  cld
  clc
  mov ax,cs
  mov ds,ax
  mov es,ax
  
  mov si,INTERVAL
  call print
  sti
  xor ax,ax
  int 1Ah
  
key_wait_loop:
  push dx
  mov ax,0100h
  int 16h
  jnz short key_pressed
  xor ax,ax
  int 1Ah
  mov cx,dx
  pop dx
  sub cx,dx
  cmp cx,18*KEY_WAIT_SEC
  ja short key_timeout
  hlt
  jmp short key_wait_loop
  
key_timeout:
  xor bx,bx
  mov ax,0E0Dh
  int 10h
  mov cx,30d
key_erase_line:
  mov ax,0E20h
  int 10h
  loop key_erase_line
  mov ax,0E0Dh
  int 10h
  xor ax,ax
  mov word[cs:0],ax
  int 19h
  
key_pressed:
  xor ax,ax
  int 16h
  pop dx
  cmp ah,3Bh
  jne key_wait_loop
  cli
  xor ax,ax
  mov ss,ax
  mov sp,0f00h
  jmp short install_int21h
  
chrout:
  cmp al,9
  je chrout_tab
  mov ah,0Eh
  int 10h
  ret
chrout_tab:
  push cx
  mov cx,6
  mov al,' '
chrout_tab_loop:
  call chrout
  loop chrout_tab_loop
  pop cx
  ret
  
print:
  mov ah,0Eh
  lodsb
  cmp al,0
  je print_end
  push si
  push bp
  int 10h
  pop bp
  pop si  
  jmp print
print_end:
  ret
   
install_int21h:
  push ds
  xor ax,ax
  mov ds,ax
  mov word[ds:20h*4],reboot_handler
  mov word[ds:20h*4+2],cs
  mov word[ds:21h*4],int21_handler
  mov word[ds:21h*4+2],cs
  pop ds
  
; binary to load is .COM by default, comment out one or the other
  ;jmp short run_exe
  jmp short run_com
  
run_com:
  push cs
  mov si,reboot_handler
  push si
  mov si,binary_to_load
  mov ax,100h
  mov es,ax
  push es
  mov al,90h
  xor di,di
  push di
  mov cx,100h
  rep stosb
  xor cx,cx
  sub cx,si
  rep movsb
  push es
  pop ds
  sti
  retf
   
run_exe:
  mov ax,ds
  mov si,binary_to_load
  add ax,[ds:si+08h]
  mov cx,[ds:si+06h]
  mov bx,[ds:si+18h]
  jcxz relocate_done
  
relocate_loop:
  mov di,[ds:bx]
  mov dx,[ds:bx+2]
  add dx,ax
  push ds
  mov ds,dx
  add [ds:di],ax
  pop ds
  add bx,4
  loop relocate_loop
  
relocate_done:
  mov bx,ax
  add bx,[ds:si+0Eh]
  mov ss,bx
  mov sp,[ds:si+10h]
  add ax,[ds:si+16h]
  push ax
  mov ax,[ds:si+14h]
  push ax
  sti
  retf
   
int21_handler:
  cli
  clc
  cmp ah,0
  je reboot_handler
  cmp ah,1
  je int21_01
  cmp ah,2
  je int21_02
  cmp ah,7
  je int21_07
  cmp ah,8
  je int21_07
  cmp ah,9
  je int21_09
  cmp ah,19h
  je int21_19
  cmp ah,25h
  je int21_25
  cmp ah,30h
  je int21_30
  cmp ah,35h
  je int21_35
  cmp ah,4Ch
  je reboot_handler
  jmp int21_error
int21_01:
  sti
  xor ax,ax
  int 16h
  call chrout
  jmp int21_exit
int21_02:
  push ax
  mov al,dl
  call chrout
  pop ax
  jmp int21_exit
int21_07:
  sti
  xor ax,ax
  int 16h
  jmp int21_exit 
int21_09:
  push bx
  mov bx,dx
int21_09_next:
  mov al,byte[es:bx]
  cmp al,'$'
  je int21_09_done
  call chrout
  inc bx
  jmp int21_09_next
int21_09_done:
  pop bx
  mov al,'$'
  jmp int21_exit
int21_19:
  mov al,0
  jmp int21_exit
int21_25:
  push cx
  mov cl,2
  shl ax,cl
  pop cx
  push si
  push bx
  push es
  mov si,ax
  xor bx,bx
  mov es,bx
  mov word[es:si],dx
  mov bx,ds
  mov word[es:si+2],bx
  pop es
  pop bx
  pop si
  jmp int21_exit
int21_30:
  mov al,3
  mov ah,3
  jmp int21_exit
int21_35:
  push ds
  push si
  xor ah,ah
  push cx
  mov cl,2
  shl ax,cl
  pop cx
  mov si,ax
  xor bx,bx
  mov ds,bx
  mov bx,word[ds:si+2]
  push bx
  mov bx,word[ds:si]
  pop es
  pop si
  pop ds
  jmp int21_exit
int21_error:
  stc
int21_exit:
  sti
  iret
  
binary_to_load:
  times ROM_SIZE*1024-13-($-$$) nop

reboot_handler:
  xor ax,ax
  mov ds,ax
  mov [ds:0472h],ax
  jmp far 0ffffh:0h
  nop